package de.hsbo.fbv.bmg.graphics.transforms;

import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;

import Jama.Matrix;

/**
 * Klasse zur Abbildung mit einer ebenen bilinearen Transformation der Form:
 * [x']   [m00 m01 m02 m03] [ x]   [m00*x + m01*y + m02 + m03*xy]
 * [y'] = [m10 m11 m12 m13]*[ y] = [m10*x + m11*y + m12 + m13*xy]
 * [1 ]   [ 0   0   1   0 ] [ 1]   [             1              ]
 * [xy]   [ 0   0   0   1 ] [xy]   [             xy             ]
 * Diese Transformation kann verkettet werden. Die Teiltransformationen sind 
 * in der richtigen Reihenfolge anzugeben, da die entsprechenden Matrizen von 
 * Links multipliziert werden.
 * 
 * @author A.Zimmermann, HS Bochum, 2011
 *
 */

public class BilinearTransform extends Transform2D {
	
	private static double eps = 1.e-15;
	
	double[][] a = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
	BilinearTransform inverse = null;
	boolean changed = false;

	/**
	 * Standardkonstruktor
	 */
	public BilinearTransform() {
		super();
	}
	
	/**
	 * Erzeugt eine bilineare Transformation mit (allen) 8 Parametern.
	 * @param a00
	 * @param a10
	 * @param a01
	 * @param a11
	 * @param a02
	 * @param a12
	 * @param a03
	 * @param a13
	 */
	public BilinearTransform(double a00, double a10, double a01, double a11,
			double a02, double a12, double a03, double a13) {
		this.a = new double[][] {
				{a00, a01, a02, a03}, {a10, a11, a12, a13},
				{0, 0, 1, 0},{0, 0, 0, 1}};
		this.changed=true;
	}
	
	/**
	 * Erzeugt eine bilineare Transformation mit den 6 Parametern einer 
	 * affinen Transformation.
	 * @param a00
	 * @param a10
	 * @param a01
	 * @param a11
	 * @param a02
	 * @param a12
	 */
	public BilinearTransform(double a00, double a10, double a01, double a11,
			double a02, double a12) {
		this.a = new double[][] {
				{a00, a01, a02, 0}, {a10, a11, a12, 0},
				{0, 0, 1, 0},{0, 0, 0, 1}};
		this.changed=true;
	}
	
	/**
	 * Erzeugt eine bilineare Transformation mit den 4 Parametern einer 
	 * homogenen Transformation.
	 * @param a00
	 * @param a10
	 * @param a02
	 * @param a12
	 */
	public BilinearTransform(double a00, double a10, double a02, double a12) {
		this.a = new double[][] {
				{a00, -a10, a02, 0}, {a10, a00, a12, 0},
				{0, 0, 1, 0},{0, 0, 0, 1}};
		this.changed=true;
	}
	
	/**
	 * Erzeugt aus einem AffineTransform-Objekt eine bilineare Transformation.
	 * @param at
	 * @return
	 */
	public static BilinearTransform create(AffineTransform at) {
		double[] a = new double[6];
		at.getMatrix(a);
		return new BilinearTransform(a[0], a[1], a[2], a[3], a[4], a[5]);
	}
	
	
	/**
	 * Fgt dieser Transformation (this) die Teiltransformation trans hinzu.
	 * Anders als bei AffineTransform sind aber die Teiltransformationen 
	 * in der richtigen Reihenfolge anzugeben, da die entsprechenden Matrizen 
	 * von Links multipliziert werden. Die Methode wird intern fr verwendet.
	 * @param trans
	 */
	public void concatenate(BilinearTransform trans) {
		Matrix mtrans = new Matrix(trans.a);
		Matrix mthis = new Matrix(this.a);
		Matrix result = mtrans.times(mthis);
		this.a = result.getArray();
		this.changed=true;
	}
	
	/**
	 * Fgt dieser Transformation (this) eine Translation hinzu.
	 * @param tx Verschiebung in x-Richtung
	 * @param ty Verschiebung in y-Richtung
	 */
	public void translate(double tx, double ty) {
		BilinearTransform trans = new BilinearTransform(
				new double[][] {
						{1,0,tx,0},{0,1,ty,0},
						{0,0,1,0},{0,0,0,1}});
		this.concatenate(trans);
	}
	
	/**
	 * Fgt dieser Transformation (this) eine Skalierung in x u. y hinzu.
	 * @param mx Mastabsfaktor der x-Achse
	 * @param my Mastabsfaktor der y-Achse
	 */
	public void scale(double mx, double my) {
		BilinearTransform trans = new BilinearTransform(
				new double[][] {
						{mx,0,0,0},{0,my,0,0},
						{0,0,1,0},{0,0,0,1}});
		this.concatenate(trans);
	}

	/**
	 * Fgt dieser Transformation (this) eine Rotation in x u. y hinzu.
	 * @param rx Drehwinkel der x-Achse in [rad]
	 * @param ry Drehwinkel der y-Achse in [rad]
	 */
	public void rotate(double rx, double ry) {
		BilinearTransform trans = new BilinearTransform(
				new double[][] {
						{Math.cos(rx),-Math.sin(rx),0,0},
						{Math.sin(ry),Math.cos(ry),0,0},
						{0,0,1,0},{0,0,0,1}});
		this.concatenate(trans);
	}
	
	/**
	 * Fgt dieser Transformation (this) eine Rotation um (x,y) hinzu.
	 * @param r Drehwinkel in [rad]
	 * @param x x-Koordinate des Drehpunktes 
	 * @param y y-Koordinate des Drehpunktes
	 */
	public void rotate(double r, double x, double y) {
		this.translate(-x, -y);
		this.rotate(r, r);
		this.translate(x, y);		
	}
	
	/**
	 * Fgt dieser Transformation (this) eine Scherung in x u. y hinzu.
	 * @param sx Schwerwert der x-Achse
	 * @param sy Schwerwert der y-Achse
	 */
	public void shear(double sx, double sy) {
		BilinearTransform trans = new BilinearTransform(
				new double[][] {
						{1,sx,0,0},{sy,1,0,0},
						{0,0,1,0},{0,0,0,1}});
		this.concatenate(trans);
	}
	
	/**
	 * Fgt dieser Transformation (this) eine bilineare Trafo hinzu.
	 * @param bx Konvergenzfaktor der x-Achse
	 * @param by Konvergenzfaktor der y-Achse
	 */
	public void converge(double bx, double by) {
		BilinearTransform trans = new BilinearTransform(
				new double[][] {
						{1,0,0,bx},{0,1,0,by},
						{0,0,1,0},{0,0,0,1}});
		this.concatenate(trans);
	}
	
	/**
	 * Gibt die Parameter einer affinen Transformation zurck. Das Objekt 
	 * wird auch dann erzeugt, wenn weitere von Null verschiedene Parameter 
	 * bestehen. 
	 * @return
	 */
	public AffineTransform toAffineTransform() {
		AffineTransform at = new AffineTransform();
		at.setTransform(a[0][0], a[1][0], a[0][1], 
						a[1][1], a[0][2], a[1][2]);
		return at;
	}
	
	@Override
	public double[] transform(double x, double y) {
		double[] dst = new double[2];
		dst[0] = a[0][0] * x + a[0][1] * y + a[0][2] + a[0][3] * x * y;
		dst[1] = a[1][0] * x + a[1][1] * y + a[1][2] + a[1][3] * x * y;
		return dst;
	}

	@Override
	public double[] inverseTransform(double x, double y) {
		// Falls nur affine Parameter vorhanden -> return
		if(Math.abs(a[0][3])==0 && Math.abs(a[1][3])==0) {
			try {
				AffineTransform at = this.toAffineTransform();
				double[] srcPts = {x,y};
				double[] dstPts = new double[2];
				at.inverseTransform(srcPts, 0, dstPts, 0, 1);
				return dstPts;
			} catch (NoninvertibleTransformException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
				return null;
			}
		}
		// sonst ... (bilineare Anteile vorhanden) 
		else {
			return inverseTransform2(x,y);
		}
	}
		
	/**
	 * Alternative Lsung zum Listing im Buch; der Ansatz beruht auf dem Newtonschen 
	 * Nherungsverfahren, bei dem ebenfalls iterativ vorgegangen wird. Die Gleichung lautet: <br>
	 * x = f(u,v) = a02 + a00*u + a01*v + a03*u*v <br>
	 * y = g(u,v) = a12 + a10*u + a11*v + a13*u*v <br>
	 * Darin sind x,y sowie die Parameter a00 bis a13 bekannt und u,v gesucht. Dafr gilt: <br>
	 * x = x0 + dx; x0 = f(u0,v0); u = u0 + du <br>
	 * y = y0 + dy; y0 = g(u0,v0); v = v0 + dv <br>
	 * Ausgedrckt als Reihenentwicklung, die nach dem ersten Glied abbricht, ergibt sich: <br>
	 * x = f(u0,v0) + (df(u0,v0)/du) * du + (df(u0,v0/dv) * dv <br>
	 * y = g(u0,v0) + (dg(u0,v0)/du) * du + (dg(u0,v0/dv) * dv <br>
	 * In jedem Iterationsschritt ist dann folgendes Gleichungssystem nach du und dv zu lsen: <br>
	 * x-x0 = dx = (a00+a03*v0) * du + (a01+a03*u0) * dv <br>
	 * y-y0 = dy = (a10+a13*v0) * du + (a11+a13*u0) * dv <br>
	 * Damit ergeben sich in jeder Iteration neue Wert fr u=u0+du und v=v0+dv, aus denen 
	 * korrigierte Werte fr x0 = f(u0+du,v0+dv) sowie y0 = g(u0+du,v0+dv) hervorgehen.
	 * Das Iterationsverfahren endet dann, wenn dx=0 und dy=0 werden. 
	 * Da vorab nichts anderes bekannt ist, startet der erste Iterationsschritt mit u0=0 und v0=0.
	 * @param x
	 * @param y
	 * @return {u,v}
	 */
	public double[] inverseTransform2(double x, double y) {
		// Anfangswerte fr die erste Iteration setzen
		double dx = x-a[0][2], dy = y-a[1][2];
		double u=0, du=0, v=0, dv=0;
		double s = Math.sqrt(x*x+y*y);
		int loop = 0;
		// Iteration
		while((Math.sqrt(dx*dx+dy*dy)/s)>eps) {
			// Berechnung der Ableitungsmatrix 
			double[][] delta = {
					{ a[0][0] + a[0][3] * v, a[0][1] + a[0][3] * u },
					{ a[1][0] + a[1][3] * v, a[1][1] + a[1][3] * u } 
			};
			// Verwendung von Jama 
			Matrix mDelta = new Matrix(delta);
			Matrix mDxy = new Matrix(new double[][] { { dx }, { dy } });
			Matrix mDuv = mDelta.solve(mDxy);
			double[] duv = mDuv.getColumnPackedCopy();
			u = u + duv[0];
			v = v + duv[1];
			// Testausgabe: Anzahl der erforderlichen Iterationen 
			System.out.println(loop++ +": "+duv[0] + " "+ duv[1]);
			// Neue Korrekturwerte berechnen
			double[] xy0 = this.transform(u, v);
			dx = x - xy0[0]; 
			dy = y - xy0[1];
		}
		return new double[] {u,v};
	}
	
	/**
	 * Lsung gem Listing im Buch 
	 * @param x
	 * @param y
	 * @return
	 */
	public double[] inverseTransform1(double x, double y) {
		// inverse Matrix der affinen Transformation beschaffen
		if(this.changed==true) {
			inverseMatrix();			
		}
		// Affine Rcktransformation rechnen
		double[] src = inverse.transform(x, y);
		// Falls nur affine Parameter vorhanden -> return
		if(Math.abs(a[0][3])<eps && Math.abs(a[1][3])<eps)
			return src;
		// sonst ... (bilineare Anteile vorhanden) 
		else {
			double u = src[0], v = src[1];
			boolean flag = true;
			// Testausgabe: Anzahl der erforderlichen Iterationen 
			int loop = 0; 
			do {
				double[] dst = transform(u, v);
				double x0 = dst[0], y0 = dst[1];
				double dx = x - x0, dy = y - y0;
				double[][] delta = {
						{ a[0][0] + a[0][3] * v, a[0][1] + a[0][3] * u },
						{ a[1][0] + a[1][3] * v, a[1][1] + a[1][3] * u } };
				// Verwendung von Jama 
				Matrix mDelta = new Matrix(delta);
				Matrix mDxy = new Matrix(new double[][] { { dx }, { dy } });
				Matrix mDuv = mDelta.solve(mDxy);
				double[] duv = mDuv.getColumnPackedCopy();
				double du = duv[0], dv = duv[1];
				// Testausgabe: Anzahl der erforderlichen Iterationen 
				System.out.println(loop++ +": "+u + " "+ v);
				u = u + du;
				v = v + dv;
				flag = Math.abs(du/u)>eps || Math.abs(dv/v)>eps;
			} while (flag);
			return (new double[] {u,v});
		}	
	}
	
	@Override
	public String toString() {
		String s = "";
		for(int i=0; i<this.a.length; i++)
			s = s+"["+rnd(a[i][0])+" "+rnd(a[i][1])+" "+rnd(a[i][2])+" "+rnd(a[i][3])+"]\n";
		return s;
	}
	
	/**
	 * Erzeugt eine bilineare Transformation mit dem Array einer 4x4-Matrix.
	 * @param a
	 */
	private BilinearTransform(double[][] a) {
		this.a = a;
		this.changed=true;
	}
	
	private double rnd(double x) {
		return Math.rint(x/eps)*eps;
	}
	
	private void inverseMatrix() {
		Matrix A = new Matrix(this.a);
		Matrix Ac = new Matrix(A.getArrayCopy());
		Ac.set(0, 3, 0);
		Ac.set(1, 3, 0);
		Matrix INV = Ac.inverse();
		this.inverse = new BilinearTransform(INV.getArray());
		this.changed = false;
	}
}
