package de.hsbo.fbv.bmg.graphics.transforms;

import java.awt.Dimension;
import java.awt.geom.AffineTransform;
import java.awt.geom.Path2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.Map;
import java.util.Set;

import Jama.Matrix;

/**
 * Factory-Klasse fr verschiedene Transformationen
 * @author A.Zimmermann, HS Bochum, 2011
 *
 */
public class TransformFactory {
	public static double DeviceFrame = 10;
	
	/**
	 * Berechnet eine ebene Transformation vom Weltkoordinatensystem ins Gertekoordinatensystem.
	 * @param world 	Rechteckiger Ausschnitt des Weltkoordinatensystems 
	 * @param device	Gre des Darstellungsfensters (width, height)
	 * @return			Transformationsobjekt vom Typ AffineTransform
	 */
	public static AffineTransform createSimpleWorldToDevice(Rectangle2D world, Dimension device) {
		try {
			// Berechnung der Transformationsparameter
			// 1. Schritt: Rcktranslation zum Mittelpunkt des Weltkoordinatenfensters
			double tx = world.getCenterX();
			double ty = world.getCenterY();
			// 2. Schritt: Skalierung, Spiegelung
			double mx = (device.getWidth()-2*DeviceFrame)/world.getWidth();
			double my = (device.getHeight()-2*DeviceFrame)/world.getHeight();
			double m = Math.min(my, mx);
			// 3. Schritt: Translation zum Mittelpunkt des Anzeigefensters
			double wx = device.getWidth()/2;
			double wy = device.getHeight()/2;
			// Anwendung der Parameter auf das Transformationsobjekt
			AffineTransform at = new AffineTransform();
			// 3. Schritt
			at.translate(wx, wy);
			// 2. Schritt
			at.scale(m, -m);
			// 1. Schritt
			at.translate(-tx, -ty);
			// Transformationsobjekt zurckgeben
			return at;
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Berechnet eine ebene Transformation vom Weltkoordinatensystem ins Gertekoordinatensystem.
	 * @param world 	Rechteckiger Ausschnitt des Weltkoordinatensystems
	 * @param alpha		Drehwinkel des Weltkoordinatensystems (mathematische Drehrichtung)
	 * @param device	Gre des Darstellungsfensters (width, height)
	 * @return			Transformationsobjekt vom Typ AffineTransform
	 */	
	public static AffineTransform createSimpleWorldToDevice(Rectangle2D world, double alpha, Dimension device) {
		try {
			// Berechnung der Transformationsparameter
			// 1. Schritt: Rcktranslation zum Mittelpunkt des Weltkoordinatenfensters
			double tx = world.getCenterX();
			double ty = world.getCenterY();
			// 2. Drehung des Weltkoordinatenfensters
			AffineTransform rotation = new AffineTransform();
			rotation.rotate(alpha);
			Path2D worldPath = new Path2D.Double(world, rotation);
			world = worldPath.getBounds2D();
			// 3. Schritt: Skalierung, Spiegelung
			double mx = (device.getWidth()-2*DeviceFrame)/world.getWidth();
			double my = (device.getHeight()-2*DeviceFrame)/world.getHeight();
			double m = Math.min(my, mx);
			// 4. Schritt: Translation zum Mittelpunkt des Anzeigefensters
			double wx = device.getWidth()/2;
			double wy = device.getHeight()/2;
			// Anwendung der Parameter auf das Transformationsobjekt
			AffineTransform at = new AffineTransform();
			// 4. Schritt
			at.translate(wx, wy);
			// 3. Schritt
			at.scale(m, -m);
			// 2. Schritt
			at.rotate(alpha);
			// 1. Schritt
			at.translate(-tx, -ty);
			// Transformationsobjekt zurckgeben
			return at;
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * Berechnet die vier Parameter einer orthogonalen Transformation aus identischen Punkten.<br>
	 * [x']   [m00 -m10 m02] [ x]   [m00*x - m10*y + m02]<br>
	 * [y'] = [m10  m00 m12]*[ y] = [m10*x + m00*y + m12]<br>
	 * [1 ]   [ 0   0   1  ] [ 1]   [         1         ]<br>
	 * 
	 * Fr die eindeutige Lsung werden zwei identische Punkte bentigt, die im Map-Objekt points
	 * zu bergeben sind. Enthlt die Map mehr Punkte, erfolgt die Berechnung nach der 
	 * Methode der kleinsten Quadrate (Ausgleichung).<br>
	 * 
	 * Das Gleichungssystem zur Auflsung der gesuchten Parameter lautet fr einen identischen Punkt P(j)
	 * mit den beiden Koordinaten x(j),y(j) bzw. x'(j),y'(j) :<br>
	 * x'(j) = x(j)*m00 - y(j)*m10 + 1*m02 + 0*m12<br>
	 * y'(j) = y(j)*m00 + x(j)*m10 + 0*m02 + 1*m12<br>
	 * :::<br>
	 * 
	 * Das Map-Objekt points enthlt die zueinander identischen Punkte in der Key-Value-Struktur
	 * (x,y):Point2D = (x',y'):Point2D<br>
	 * 
	 * Um die berechneten Transformationsparameter und (im Fall einer mehrdeutigen Lsung) 
	 * ihre statistischen Genauigkeiten ausgeben zu knnen, erwartet die Methode als bergabe- 
	 * parameter eine weitere leere Map, in die die Ergebnisse der Berechnung eingefgt werden.
	 * Sie hat die Key-Value-Struktur Parameter:String = Wert:Number<br>
	 * 
	 * @param points - Map der identischen Punkte
	 * @return Transformationsobjekt vom Typ BilinearTransform
	 */
	public static BilinearTransform create4Parameter(Map<Point2D,Point2D> points) {
		// Fallunterscheidung
		int npts = points.size();
		try {
			if(npts<2) {
				throw new Exception("Number of points less than two!");
			}
			else {
				// Aufstellen der Koeffizientenmatrix und des Absolutvektors
				double[][] x = new double[npts*2][4];
				double[] y   = new double[npts*2];
				Set<Point2D> keys = points.keySet();
				Point2D[] pa = new Point2D[npts];
				keys.toArray(pa);
				for(int i=0; i<npts; i++) {
					// Koeffizienten
					x[2*i+0][0]	=  pa[i].getX(); 	//  m00 in x'
					x[2*i+0][1]	= -pa[i].getY(); 	// -m10 in x'
					x[2*i+0][2]	=  1;				//  m02 in x'
					x[2*i+0][3]	=  0;				//	m12 in x'
					x[2*i+1][0]	=  pa[i].getY(); 	//  m00 in y'
					x[2*i+1][1]	=  pa[i].getX(); 	//  m10 in y'
					x[2*i+1][2]	=  0;				//  m02 in y'
					x[2*i+1][3]	=  1;				//	m12 in y'
					// Absolutwerte
					Point2D pe 	= (Point2D)points.get(pa[i]);
					y[2*i+0]	= pe.getX();		// x'
					y[2*i+1]	= pe.getY();		// y'
				}
				// Gleichungssystem lsen mit Jama
				double[] z = solveEquations(x, y);
				return new BilinearTransform(z[0], z[1], z[2], z[3]);
			}
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * Berechnet die sechs Parameter einer affinen Transformation aus identischen Punkten.<br>
	 * [x']   [m00 m01 m02] [ x]   [m00*x + m01*y + m02]<br>
	 * [y'] = [m10 m11 m12]*[ y] = [m10*x + m11*y + m12]<br>
	 * [1 ]   [ 0   0   1 ] [ 1]   [         1         ]<br>
	 * 
	 * Fr die eindeutige Lsung werden drei identische Punkte bentigt, die im Map-Objekt points
	 * zu bergeben sind. Enthlt die Map mehr Punkte, erfolgt die Berechnung nach der 
	 * Methode der kleinsten Quadrate (Ausgleichung).<br>
	 * 
	 * Das Gleichungssystem zur Auflsung der gesuchten Parameter lautet fr einen identischen Punkt P(j)
	 * mit den beiden Koordinaten x(j),y(j) bzw. x'(j),y'(j) :<br>
	 * x'(j) = x(j)*m00 +    0*m10 + y(j)*m01 +    0*m11 + 1*m02 + 0*m12<br>
	 * y'(j) =    0*m00 + x(j)*m10 +    0*m01 + y(j)*m11 + 0*m02 + 1*m12<br>
	 * :::<br>
	 * 
	 * Das Map-Objekt points enthlt die zueinander identischen Punkte in der Key-Value-Struktur
	 * (x,y):Point2D = (x',y'):Point2D<br>
	 * 
	 * Um die berechneten Transformationsparameter und (im Fall einer mehrdeutigen Lsung) 
	 * ihre statistischen Genauigkeiten ausgeben zu knnen, erwartet die Methode als bergabe- 
	 * parameter eine weitere leere Map, in die die Ergebnisse der Berechnung eingefgt werden.
	 * Sie hat die Key-Value-Struktur Parameter:String = Wert:Number<br>
	 * 
	 * @param points - Map der identischen Punkte
	 * @return Transformationsobjekt vom Typ BilinearTransform
	 */
	public static BilinearTransform create6Parameter(Map<Point2D,Point2D> points) {
		// Fallunterscheidung
		int npts = points.size();
		try {
			if(npts<3) {
				throw new Exception("Number of points less than three!");
			}
			else {
				// Aufstellen der Koeffizientenmatrix und des Absolutvektors
				double[][] x = new double[npts*2][6];
				double[] y   = new double[npts*2];
				Set<Point2D> keys = points.keySet();
				Point2D[] pa = new Point2D[npts];
				keys.toArray(pa);
				for(int i=0; i<npts; i++) {
					// Koeffizienten
					x[2*i+0][0]	=  pa[i].getX(); 	//  m00 in x'
					x[2*i+0][1]	=  0; 				//  m10 in x'
					x[2*i+0][2]	=  pa[i].getY(); 	//  m01 in x'
					x[2*i+0][3]	=  0; 				//  m11 in x'
					x[2*i+0][4]	=  1;				//  m02 in x'
					x[2*i+0][5]	=  0;				//	m12 in x'
					x[2*i+1][0]	=  0; 				//  m00 in y'
					x[2*i+1][1]	=  pa[i].getX(); 	//  m10 in y'
					x[2*i+1][2]	=  0; 				//  m01 in y'
					x[2*i+1][3]	=  pa[i].getY(); 	//  m11 in y'
					x[2*i+1][4]	=  0;				//  m02 in y'
					x[2*i+1][5]	=  1;				//	m12 in y'
					// Absolutwerte
					Point2D pe 	= (Point2D)points.get(pa[i]);
					y[2*i+0]	= pe.getX();		// x'
					y[2*i+1]	= pe.getY();		// y'
				}
				// Gleichungssystem lsen mit Jama
				double[] z = solveEquations(x, y);
				return new BilinearTransform(z[0], z[1], z[2], z[3], z[4], z[5]);
			}
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * Berechnet die acht Parameter einer bilinearen Transformation aus identischen Punkten.<br>
	 * [x']   [m00 m01 m02 m03] [ x]   [m00*x + m01*y + m02 + m03*xy]<br>
	 * [y'] = [m10 m11 m12 m13]*[ y] = [m10*x + m11*y + m12 + m13*xy]<br>
	 * [1 ]   [ 0   0   1   0 ] [ 1]   [              1             ]<br>
	 * [xy]   [ 0   0   0   1 ] [xy]   [             xy             ]<br>
	 * 
	 * Fr die eindeutige Lsung werden vier identische Punkte bentigt, die im Map-Objekt points
	 * zu bergeben sind. Enthlt die Map mehr Punkte, erfolgt die Berechnung nach der 
	 * Methode der kleinsten Quadrate (Ausgleichung).<br>
	 * 
	 * Das Gleichungssystem zur Auflsung der gesuchten Parameter lautet fr einen identischen Punkt P(j)
	 * mit den beiden Koordinaten x(j),y(j) bzw. x'(j),y'(j) und p(j)=x(j)*y(j):<br>
	 * x'(j) = x(j)*m00 +    0*m10 + y(j)*m01 +    0*m11 + 1*m02 + 0*m12 + p(j)*m03 +    0*m13<br>
	 * y'(j) =    0*m00 + x(j)*m10 +    0*m01 + y(j)*m11 + 0*m02 + 1*m12 +    0*m03 + p(j)*m13<br>
	 * :::<br>
	 * 
	 * Das Map-Objekt points enthlt die zueinander identischen Punkte in der Key-Value-Struktur
	 * (x,y):Point2D = (x',y'):Point2D<br>
	 * 
	 * Um die berechneten Transformationsparameter und (im Fall einer mehrdeutigen Lsung) 
	 * ihre statistischen Genauigkeiten ausgeben zu knnen, erwartet die Methode als bergabe- 
	 * parameter eine weitere leere Map, in die die Ergebnisse der Berechnung eingefgt werden.
	 * Sie hat die Key-Value-Struktur Parameter:String = Wert:Number<br>
	 * 
	 * @param points - Map der identischen Punkte
	 * @return Transformationsobjekt vom Typ BilinearTransform
	 */
	public static BilinearTransform create8Parameter(Map<Point2D,Point2D> points) {
		// Fallunterscheidung
		int npts = points.size();
		try {
			if(npts<4) {
				throw new Exception("Number of points less than four!");
			}
			else {
				// Aufstellen der Koeffizientenmatrix und des Absolutvektors
				double[][] x = new double[npts*2][8];
				double[] y   = new double[npts*2];
				Set<Point2D> keys = points.keySet();
				Point2D[] pa = new Point2D[npts];
				keys.toArray(pa);
				for(int i=0; i<npts; i++) {
					// Koeffizienten
					double xy = pa[i].getX()*pa[i].getY();
					x[2*i+0][0]	=  pa[i].getX(); 	//  m00 in x'
					x[2*i+0][1]	=  0; 				//  m10 in x'
					x[2*i+0][2]	=  pa[i].getY(); 	//  m01 in x'
					x[2*i+0][3]	=  0; 				//  m11 in x'
					x[2*i+0][4]	=  1;				//  m02 in x'
					x[2*i+0][5]	=  0;				//	m12 in x'
					x[2*i+0][6]	=  xy;				//  m03 in x'
					x[2*i+0][7]	=  0;				//	m13 in x'
					x[2*i+1][0]	=  0; 				//  m00 in y'
					x[2*i+1][1]	=  pa[i].getX(); 	//  m10 in y'
					x[2*i+1][2]	=  0; 				//  m01 in y'
					x[2*i+1][3]	=  pa[i].getY(); 	//  m11 in y'
					x[2*i+1][4]	=  0;				//  m02 in y'
					x[2*i+1][5]	=  1;				//	m12 in y'
					x[2*i+1][6]	=  0;				//  m03 in y'
					x[2*i+1][7]	=  xy;				//	m13 in y'
					// Absolutwerte
					Point2D pe 	= (Point2D)points.get(pa[i]);
					y[2*i+0]	= pe.getX();		// x'
					y[2*i+1]	= pe.getY();		// y'
				}
				// Gleichungssystem lsen mit Jama
				double[] z = solveEquations(x, y);
				return new BilinearTransform(z[0], z[1], z[2], z[3], z[4], z[5], z[6], z[7]);
			}
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * Lst das Gleichungssystem mit Hilfe der Matrix-Bibliothek jama.jar
	 * @param x
	 * @param y
	 * @param v
	 * @return
	 */
	private static double[] solveEquations(double[][] x, double[] y) {
		Matrix X = new Matrix(x);
		double[][] my = new double[y.length][1];
		for(int i=0; i<y.length; i++)
			my[i][0] = y[i];
		Matrix Y = new Matrix(my);
		Matrix Z = X.solve(Y);
		double[] z = Z.getColumnPackedCopy();
		return z;
	}
	
	/**
	 * Berechnet die Parameter fr eine perpektivische Transformation und gibt sie als
	 * Transform2hD-Objekt zurck.
	 * @param posx x-Wert Drehpunkt des Modells im Weltkoordinatensystem  
	 * @param posy y-Wert
	 * @param posz z-Wert
	 * @param dist Abstand der Kamera
	 * @param decl Neigung der Aufnahmeachse (90 horizontal, 0 vertikal) 
	 * @param azim Azimuth der Aufnahmeachse (Norden 0, Osten 90 usw.) 
	 * @return
	 */
	public static Transform2hD createPerspective(double posx, double posy, double posz,
			double dist, double decl, double azim) {
		Transform2hD result = new Transform2hD();
		result.translate(-posx, -posy, -posz);
		result.rotateZ(Math.toRadians(azim));
		result.rotateX(Math.toRadians(-decl));
		result.translate(0, 0, -dist);
		result.addProjection(Transform2hD.PROJ_OPTICAL);
		return result;
	}
	
	/**
	 * Berechnet die Parameter fr eine perpektivische Transformation und gibt sie als
	 * Transform2hD-Objekt zurck. Zugleich kann das Gelnderelief berhht dargestellt 
	 * werden.
	 * @param mh   Skalierung der Hhen des Modells
	 * @param posx x-Wert Drehpunkt des Modells im Weltkoordinatensystem
	 * @param posy y-Wert
	 * @param posz z-Wert
	 * @param dist Abstand der Kamera
	 * @param decl Neigung der Aufnahmeachse (90 horizontal, 0 vertikal) 
	 * @param azim Azimuth der Aufnahmeachse (Norden 0, Osten 90 usw.) 
	 * @return
	 */
	public static Transform2hD createPerspective(double mh, double posx, double posy, double posz, 
			double dist, double decl, double azim) {
		Transform2hD result = new Transform2hD();
		result.scale(1, 1, mh);
		result.translate(-posx, -posy, -posz);
		result.rotateZ(Math.toRadians(azim));
		result.rotateX(Math.toRadians(-decl));
		result.translate(0, 0, -dist);
		result.addProjection(Transform2hD.PROJ_OPTICAL);
		return result;
	}
	
	/**
	 * Berechnet die Parameter fr eine perpektivische Transformation und gibt sie als
	 * Transform2hD-Objekt zurck. Zugleich kann das Gelnderelief berhht dargestellt 
	 * werden.
	 * @param mh   Skalierung der Hhen des Modells
	 * @param posx x-Wert Drehpunkt des Modells im Weltkoordinatensystem
	 * @param posy y-Wert
	 * @param posz z-Wert
	 * @param dist Abstand der Kamera
	 * @param decl Neigung der Aufnahmeachse (90 horizontal, 0 vertikal) 
	 * @param azim Azimuth der Aufnahmeachse (Norden 0, Osten 90 usw.) 
	 * @return
	 */
	public static Transform2hD[] createPseudo3D(double mh, double baseline, 
			double posx, double posy, double posz, 
			double dist, double decl, double azim) {
		// left
		Transform2hD left = new Transform2hD();
		left.scale(1, 1, mh);
		left.translate(-posx, -posy, -posz);
		left.rotateZ(Math.toRadians(azim));
		left.rotateX(Math.toRadians(-decl));
		left.translate(baseline/2, 0, 0);
		left.translate(0, 0, -dist);
		left.addProjection(Transform2hD.PROJ_OPTICAL);
		// right
		Transform2hD right = new Transform2hD();
		right.scale(1, 1, mh);
		right.translate(-posx, -posy, -posz);
		right.rotateZ(Math.toRadians(azim));
		right.rotateX(Math.toRadians(-decl));
		right.translate(-baseline/2, 0, 0);
		right.translate(0, 0, -dist);
		right.addProjection(Transform2hD.PROJ_OPTICAL);
		
		return new Transform2hD[] {left, right};
	}
	
	
}
