package de.hsbo.fbv.bmg.graphics.viewer.commands;

import java.awt.Graphics;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import javax.swing.JFrame;
import javax.swing.JPanel;

/**
 * Schnittstelle Map zur Anzeige von Geometrien in einem Fenster am Desktop.
 * Die Schnittstelle setzt voraus, dass sie mit den Klassen des Pakets 
 * javax.swing implementiert wird, da sich verschiedene Methoden darauf beziehen.
 * Die Schnittstelle enthlt dynamisch einsetzbare Befehle, die nach dem 
 * Entwurfsmuster <b>Command</b> gestaltet sind.
 * @author zimmermann
 * @see MapAdapter
 */
public interface Map {
	
	/**
	 * Verfgbare Befehle (Commands) mit denen das Verhalten der Map gestaltet 
	 * wird. - Sie wirken wie die Tasten einer Fernbedienung, die dynamisch 
	 * belegt werden knnen.
	 */
	public static final int CmdPaint 			=  0;
	public static final int CmdMouseReleased	=  1;
	public static final int CmdMousePressed 	=  2;
	public static final int CmdMouseExited		=  3;
	public static final int CmdMouseEntered 	=  4;
	public static final int CmdMouseClicked		=  5;
	public static final int CmdMouseMoved 		=  6;
	public static final int CmdMouseDragged		=  7;
	public static final int CmdMouseWheelMoved	=  8;
	public static final int CmdKeyTyped 		=  9;
	public static final int CmdKeyReleased		= 10;
	public static final int CmdKeyPressed 		= 11;
	
	/**
	 * Zentrale Methode, mit der die Cammands und somit 
	 * das Verhalten der Map festgelegt wird.
	 * @param command
	 */
	public void initialize(Command command);
	
	/**
	 * Stellt den Kontakt zum steuernden Programm (Controller) her.
	 * @param controller
	 */
	public void setController(Controller controller);
	
	/**
	 * Gibt den Controller zurck.
	 * @return
	 */
	public Controller getController();
	
	/**
	 * Gibt den Weltkoordinatenausschnitt zur Anzeige an.
	 * @return
	 */
	public Rectangle2D getWorldBounds();
	
	/**
	 * Gibt die Transformation zur Abbildung vom Weltkkoordinatensystem 
	 * ins Gertekoordinatensystem zurck.
	 * @return
	 */
	public AffineTransform getTransform();
	
	/**
	 * Setzt den Weltkoordinatenausschnitt zur Anzeige in der Map.
	 * @return
	 */
	public void setWorldBounds(Rectangle2D r);
	
	/**
	 * Setzt die Transformation zur Abbildung vom Weltkkoordinatensystem 
	 * ins Gertekoordinatensystem.
	 * @return
	 */
	public void setTransform(AffineTransform at);
	
	/**
	 * Berechnet die Transformation zur Abbildung vom Weltkkoordinatensystem 
	 * ins Gertekoordinatensystem.
	 * @param world Weltkkordinatenausschnitt
	 * @return
	 */
	public AffineTransform getTransform(Rectangle2D world);

	/**
	 * Gibt das javax.swing.JPanel-Objekt zurck.
	 * @return
	 */
	public JPanel getPanel();
	
	/**
	 * Gibt das javax.swing.JFrame-Objekt zurck, in das das Map-Objekt
	 * eingebettet ist.
	 * @return
	 */
	public JFrame getFrame();

	/**
	 * Setzt das javax.swing.JFrame-Objekt, in das das Map-Objekt eingebettet ist.
	 * @param frame
	 */
	public void setFrame(JFrame frame);
	
	/**
	 * Gibt das Graphics-Objekt zur Verwaltung der Anzeige des JPanels zurck.
	 * @return
	 */
	public Graphics getGraphics();
	
	/**
	 * Belegt den Befehls vom Index i mit dem passenden Befehlsobjekt. 
	 * Die Methode kann intern bei initialize() verwendet werden, oder das 
	 * Verhalten der Map dynamisch, d.h. whrend der Laufzeit verndern.
	 * @param cmd Index der 'Fernbedienungstaste'
	 * @param command Befehl gem dem Command-Entwurfsmuster
	 */
	public void setCommand(int cmd, Command command);
	
	/**
	 * Veranlasst, dass der Befehl vom Index i ausgefhrt wird. 
	 * @param cmd
	 */
	public void executeCommand(int cmd);
	
	/**
	 * Setzt den inneren Zustand der Map gem einem Zustandsautomaten. 
	 * Es wird dem Programmierer der Commands berlassen, fr die 
	 * Implementierung eines funktionierenden Zustandsautomaten zu sorgen. 
	 * Diese Methode stellt nur die notwendige Voraussetzung dazu bereit.
	 * @param state
	 * @see getState()
	 */
	public void setState(int state);
	
	/**
	 * Ruft den inneren Zustand der Map ab. Der Anfangswert ist 0 .
	 * Es wird dem Programmierer der Commands berlassen, fr die 
	 * Implementierung eines funktionierenden Zustandsautomaten zu sorgen. 
	 * Diese Methode stellt nur die notwendige Voraussetzung dazu bereit.
	 * @return
	 * @see setState(int state)
	 */
	public int getState();

	/**
	 * Ruft das KeyEvent-Objekt des letzten Ereignisses ab. 
	 * @return
	 */
	public KeyEvent getLastKeyEvent();
	
	/**
	 * Ruft das KeyEvent-Objekt des aktuellen Ereignisses ab. 
	 * @return
	 */
	public KeyEvent getActKeyEvent();
	
	/**
	 * Ruft das MouseEvent-Objekt (incl. MouseWheelEvents) des letzten Ereignisses ab. 
	 * @return
	 */
	public MouseEvent getLastMouseEvent();

	/**
	 * Ruft das MouseEvent-Objekt (incl. MouseWheelEvents) des aktuellen Ereignisses ab. 
	 * @return
	 */	
	public MouseEvent getActMouseEvent();

	/**
	 * Ruft die transformierte Koordinate des letzten Maus-Ereignisses ab. 
	 * @return
	 */
	public Point2D getLastPoint();
	
	/**
	 * Ruft die transformierte Koordinate des aktuellen Maus-Ereignisses ab. 
	 * @return
	 */
	public Point2D getActPoint();
	
	/**
	 * Zeichnet die Grpahik in der Map neu.
	 */
	public void repaint();
	
}
